;;;
;;;  parser.scm: parser generation functions
;;;
;;;
;;;  Copyright (C) 2001 Samuel Hocevar
;;;  $Id: parser.scm,v 1.3 2001/11/27 18:24:57 sam Exp $
;;; 
;;;  Authors: Samuel Hocevar <sam@zoy.org>
;;; 
;;;  This program is free software; you can redistribute it and/or modify
;;;  it under the terms of the GNU General Public License as published by
;;;  the Free Software Foundation; either version 2 of the License, or
;;;  (at your option) any later version.
;;;  
;;;  This program is distributed in the hope that it will be useful,
;;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;  GNU General Public License for more details.
;;; 
;;;  You should have received a copy of the GNU General Public License
;;;  along with this program; if not, write to the Free Software
;;;  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  parser: parse source s using grammar g

(define make-parser (lambda (g)
  (eval (grammar->parser g))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  grammar->parser: generate a parser from grammar
;;;
;;;  The inner loop is grammar->parser/in which makes a call to
;;;  production->parser for each production in the grammar.

(define grammar->parser (lambda (g)
  `(lambda (src)
    (letrec ,(grammar->parser/in g g)
      (car (,(make-symbol "rs" (caar g)) (cons () src)))))))

(define grammar->parser/in (lambda (gn g)
  (if (null? gn)
    ()
    (cons (production->parser (car gn) g)
          (grammar->parser/in (cdr gn) g)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  production->parser: generate a parser from a production
;;;
;;;  This function checks whether the production contains only lexems, and
;;;  if it does, it will try to optimize the parser generation by calling
;;;  ruleset->lexem-parser. Otherwise, ruleset->parser is called.

(define production->parser (lambda (p g)
  (if (null? (cdr p))
    ()
    (list (make-symbol "rs" (car p))
      (ruleset->parser (cdr p) (car p) #f 0 g)))))

(define ruleset->parser (lambda (s pn l? li g)
  `(lambda (s0)
     ,(cond 
       (l?
        (ruleset->multirule-parser s pn l? li g))
       ((has-ruleset-only-lexems? s)
        (ruleset->lexem-parser s pn l?))
       ((null? (cdr s))
        (ruleset->one-rule-parser s pn l? li g))
       (else
        (ruleset->multirule-parser s pn l? li g))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  ruleset->parser: generic parser generator.

(define ruleset->multirule-parser (lambda (s pn l? li g)
  `(cond ,@(ruleset->multirule-parser/in s pn l? li g))))

(define ruleset->multirule-parser/in (lambda (s pn l? li g)
  (if (null? s)
    `((else
       ,(if l?
         '(cons () (cdr s0))
         `(parse-error s0 ,(metaname->string pn)))))
    (if (null? (car s))
      (if (null? (cdr s))
        '((else (cons (list 'index) (cdr s0))))
        (ruleset->multirule-parser/in (cons (cadr s) (cons (car s) (cddr s)) l? li g)))
      (cons (rule->parser (car s) pn l? li g)
            (ruleset->multirule-parser/in (cdr s) pn l? li g))))))

(define rule->parser (lambda (r pn l? li g)
  (cond
    ; rule is empty
    ((null? r)
     `(cons (list (',pn)) (cdr s0)))
    ; first item of the rule is a string
    ((string? (car r))
     `((reco s0 ',(list (car r)))
       ,(rule->inner-parser (cdr r) 1 '((chop s0) s0)
                            (list (car r) `',pn) l? li g)))
    ; first item of the rule is a metaname
    ((symbol? (car r))
     `((reco s0 ',(get-symbol-tete (car r) g))
       (let ((s1 (,(make-symbol "rs" (car r)) s0)))
         (if (car s1)
           ,(rule->inner-parser (cdr r) 2 '(s1 s0)
                                (list '(car s1) `',pn) l? li g)))))
    ; first item of the rule is a loop
    ((list? (car r))
     (let ((loops (production->loops r pn li g))
           (newcall `("TODO !!!" ,r ,pn)))
          (if (null? loops) newcall `(letrec ,loops ,newcall))))
    ; ???
    (else
     (dtc-error "rule->parser failed, bogus grammar?")
     (display r) (newline)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  ruleset->lexem-parser: generate a parser from a ruleset containing
;;;  only terminal rules.

(define ruleset->lexem-parser (lambda (s pn l?)
  `(if (reco s0 ',(make-flat-list s))
     (cons (list ',pn (cadr s0)) (cddr s0))
     ,(if l?
        '(cons () (cdr s0))
        `(parse-error s0 ,(make-flat-string s))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  ruleset->one-rule-parser: generate a parser from a ruleset containing
;;;  only one rule.

(define ruleset->one-rule-parser (lambda (s pn l? li g)
  (rule->inner-parser (car s) 1 '(s0) `(',pn) l? li g)))

(define rule->inner-parser (lambda (r vi vl e l? li g)
  (let ((vi2 (number->var vi)))
    (cond
      ; rule is empty
      ((null? r)
       (if l?
         `(let ((,vi2 (,(make-symbol "loop" (1- li)) ,(car vl))))
            (cons (append (list ,@(cdr (retourne e))) (car ,vi2)) (cdr ,vi2)))
         `(cons (list ,@(retourne e)) (cdr ,(car vl)))))
      ; first item of the rule is a string
      ((string? (car r))
       `(if (reco ,(car vl) '(,(car r)))
         ,(rule->inner-parser (cdr r) vi (cons `(chop ,(car vl)) vl)
                              (cons (car r) e) l? li g)
         ,(if (and l? (equal? 's0 (car vl)))
           '(cons () (cdr s0))
           `(parse-error ,(car vl) ,(car r)))))
      ; first item of the rule is a metaname
      ((symbol? (car r))
       `(let ((,vi2 (,(make-symbol "rs" (car r)) ,(car vl))))
          (if (car ,vi2)
            ,(rule->inner-parser (cdr r) (1+ vi) (cons vi2 vl)
                                 (cons (list 'car vi2) e) l? li g)
            ,(if (and l? (equal? 's0 (car vl)))
               '(cons () (cdr s0))
               `(parse-error ,(car vl) ,(metaname->string (car r)))))))
      ; first item of the rule is a loop
      ((list? (car r))
       `(letrec
         ,(production->loops (car r) (caar r) li g)
         ,(cond
            ((equal? '* (caar r))
             `(let* ((,vi2 (,(make-symbol "loop" li) ,(car vl))))
               ,(rule->inner-parser (cdr r) (1+ vi) (cons vi2 vl)
                                    (cons (list 'car vi2) e) l? li g)))
            ((equal? '+ (caar r))
             '(" +  loops unimplemented yet")))))
      ; ???
      (else
       (dtc-error "rule->inner-parser failed, bogus grammar?")
       (display r) (newline))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  production->loops: generate a list of loop functions from a production
;;;
;;;  This function makes recursive calls to ruleset->classic-parser which
;;;  will generate inner loop routines.

(define production->loops (lambda (r pn li g)
  (ruleset->loops (cdr r) pn li g)))

(define ruleset->loops (lambda (s pn li g)
  (if (null? s)
    ()
    (if (list? (car s))
      (cons
        `(,(make-symbol "loop" li)
          ,(ruleset->parser s pn #t (1+ li) g))
        (ruleset->loops (cdr s) pn li g))
      (ruleset->loops (cdr s) pn li g)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;
;;;  Following functions are local
;;;

; generic EOF parser
(define rs_$ (lambda (s0)
  (if (null? (cdr s0))
    '($)
    (parse-error s0 "EOF"))))

(define retourne (lambda (l)
  (if (not l)
    #f
    (letrec ((r/ (lambda (l m) (if (null? l) m (r/ (cdr l) (cons (car l) m))))))
      (r/ l ())))))

(define reco (lambda (s lst) (if (null? (cdr s)) #f (member (cadr s) lst))))

(define chop (lambda (s) (cons (car s) (cddr s))))

(define parse-error (lambda (s n)
  (letrec
    ((nxch_n
       (lambda (l n)
         (if (null? l)
           "EOF"
           (if (zero? n)
             "..."
             (string-append (car l) " " (nxch_n (cdr l) (- n 1)))))))
     (nxch
       (lambda (s)
         (nxch_n (cdr s) 5))))
  (dtc-error "parser stuck before  " (nxch s) " , expecting  " n " ")
  (cons #f (cdr s)))))

(define number->var (lambda (n)
  (string->symbol (string-append "s" (number->string n)))))

(define metaname->string (lambda (n)
  (string-append "<" (symbol->string n) ">")))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  make-flat-list: flatten a ruleset

(define make-flat-list (lambda (s)
  (if (null? s) () (cons (caar s) (make-flat-list (cdr s))))))

(define make-flat-string (lambda (s)
  (if (null? s) "" (if (null? (cdr s)) (caar s) (string-append (caar s) "|" (make-flat-string (cdr s)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  make-symbol: return a symbol from a string and a symbol or number

(define make-symbol (lambda (s n)
  (string->symbol
    (string-append s "_" (cond ((symbol? n) (symbol->string n))
                               ((number? n) (number->string n))
                               (else "?"))))))

