;;;
;;;  codegen.scm: code generation for the BE's language
;;;
;;;
;;;  Copyright (C) 2001 Samuel Hocevar
;;;  $Id: codegen.scm,v 1.2 2001/11/27 15:35:40 sam Exp $
;;; 
;;;  Authors: Samuel Hocevar <sam@zoy.org>
;;; 
;;;  This program is free software; you can redistribute it and/or modify
;;;  it under the terms of the GNU General Public License as published by
;;;  the Free Software Foundation; either version 2 of the License, or
;;;  (at your option) any later version.
;;;  
;;;  This program is distributed in the hope that it will be useful,
;;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;  GNU General Public License for more details.
;;; 
;;;  You should have received a copy of the GNU General Public License
;;;  along with this program; if not, write to the Free Software
;;;  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.


;;;
;;;  generate-pseudocode: generate assembly code from a syntax tree
;;;

(define generate-pseudocode (lambda (tree)
  (generate-pseudocode/in (cdr tree))))

(define generate-pseudocode/in (lambda (tree)
  (if (null? tree)
    ()
    (cons (subtree-gen (car tree) 0)
          (generate-pseudocode/in (cdr tree))))))

;;;
;;;  subtree-gen: generate pseudocode from a subtree
;;;

(define subtree-gen (lambda (tree gar)
  ; What object did we meet ?
  (case (car tree)
    ; a := ...
    ((var_affect)
     (list
       (subtree-gen-and-LDA (caddr tree) gar)
       (string-append "STA " (cadadr tree))
       ""))
    ; a[...] := ...
    ((tab_affect)
     (list
       (tab-gen (cddadr tree) gar (cadadr tree) 0)
       (string-append "STA w" (number->string gar))
       (subtree-gen-and-LDA (caddr tree) (1+ gar))
       (string-append "STA @w" (number->string gar))
       ""))
    ; a
    ((var)
     (string-append "LDA " (cadr tree)))
    ; a[...]
    ((tab)
     (tab-gen (cddr tree) gar (cadr tree) 0))
    ; SI ... ALORS ... SINON ...
    ((test)
     ; TODO: we can optimize stuff here if both variables are tables
     (list
       (binary-gen (cons '? (cdadr tree)) gar)
       (caadr tree)
       (subtree-gen-and-LDA (caddr tree) gar)
       'j0
       (subtree-gen-and-LDA (cadddr tree) gar)
       'j1))
    ; -X
    ((neg)
     (list
       (subtree-gen-and-LDA (cadr tree) gar)
       "NEG"))
    ; Fast generation is done in binary-gen
    ((+ - * /)
     (binary-gen tree gar))
    (else
      (string-append "found an invalid code:" (symbol->string (car tree)))))))

;;;
;;;  Parse an object just like subtree-gen, except we put its value in the
;;;  stack (instead of its address) if it was a table.
;;;

(define subtree-gen-and-LDA (lambda (tree gar)
  (cons
    (subtree-gen tree gar)
    (if (equal? 'tab (car tree))
      (let ((s (number->string gar)))
        (list
          (string-append "STA w" s)
          (string-append "LDA @w" s)))
      ()))))

;;;
;;;  Calculate a tab address from a list of values
;;;

(define tab-gen (lambda (tree gar var i)
  (if (null? tree)
    (list (string-append "PEA " var)
          (string-append "PSH #" (number->string i))
          "JSR ADELEM")
    (cons
      (case (caar tree)
        ; X = simple variable, we do PSH X
        ((var)
          (string-append "PSH " (cadar tree)))
        ; X = tableau, on le calcule et on fait STA w0 PSH @w0
        ((tab)
          (list
            (subtree-gen (car tree) gar)
            (string-append "STA w" (number->string gar))
            (string-append "PSH @w" (number->string gar))))
        ; X = truc complexe, on le calcule et on fait STA w0 PSH w0
        (else
          (list
            (subtree-gen (car tree) gar)
            (string-append "STA w" (number->string gar))
            (string-append "PSH w" (number->string gar)))))
      (tab-gen (cdr tree) gar var (1+ i))))))

;;;
;;;  Recursively calculate :
;;;   * either a simple arithmetic expression of the form (op X Y)
;;;   * either a simple test expression of the form (? X Y)
;;;

(define binary-gen (lambda (tree gar)
  (let ((op (cadr (assoc (car tree) '((+ "ADD ")
                                      (- "SUB ")
                                      (* "MUL ")
                                      (/ "DIV ")
                                      (? "CMP ")))))
        (X (cadr tree)) (Y (caddr tree)))
    (case (car Y)
      ; If Y is a simple var, calculate X first
      ((var)
        (list
          (subtree-gen-and-LDA X gar) ; Put X in accumulator
          (string-append op (cadr Y)))) ; op Y
      ; If Y is a tab, calculate it first
      ((tab)
        (list
          (subtree-gen Y gar)
          (string-append "STA w" (number->string gar))
          (subtree-gen-and-LDA X (1+ gar))
          (string-append op "@w" (number->string gar))))
      ; If Y is a complex expression, calculate it first
      (else
        (list
          (subtree-gen Y gar)
          (string-append "STA w" (number->string gar))
          (subtree-gen-and-LDA X (1+ gar))
          (string-append op "w" (number->string gar))))))))

;;;
;;;  format-pseudocode: nicely format pseudocode
;;;  output-pseudocode: nicely print pseudocode to stdout
;;;

(define format-pseudocode (lambda (code) (fmt/in "" () code () 0)))
(define output-pseudocode (lambda (code) (fmt/in #f () code () 0)))

(define fmt/in (lambda (result elem code lstack lindex)
  (cond
    ; List ? Recursively format it
    ((list? elem)
     (if (null? elem)
       (if (null? code)
         result
         (fmt/in result (car code) (cdr code) lstack lindex))
       (fmt/in result (car elem) (cons (cdr elem) code) lstack lindex)))
    ; Handle simple commands
    ((string? elem)
     (fmt/append result (string-append "\t" elem "\n")
                 (car code) (cdr code) lstack lindex))
    ; Handle loops
    ((symbol? elem)
     (cond
       ((equal? elem 'j0)
        (fmt/append result (string-append "\tBRA $"
                             (number->string (1+ (car lstack))) "\n"
                             "$" (number->string (car lstack)) ":\n")
                    (car code) (cdr code) lstack lindex))
       ((equal? elem 'j1)
        (fmt/append result (string-append
                             "$" (number->string (1+ (car lstack))) ":\n")
                    (car code) (cdr code) (cdr lstack) lindex))
       ((member elem '(= != < > <= >=))
        (fmt/append result
          (string-append
            ; yeah, the order is reversed for BGE/BLT and BGT/BLE
            "\t" (cadr (assoc elem '((!= "BNE") (>= "BLT") (<= "BGT")
                                     (= "BEQ") (> "BLE") (< "BGE"))))
            " $" (number->string lindex) "\n")
          (car code) (cdr code) (cons lindex lstack) (+ 2 lindex)))
       (else
        (fmt/append result "dtc error: unknown symbol!\n"
                    (car code) (cdr code) lstack lindex))))
    (else
     (fmt/append result "dtc error: unknown pseudocode !\n"
                 (car code) (cdr code) lstack lindex)))))

(define fmt/append (lambda (result opcode elem code lstack lindex)
  (if result
    (fmt/in (string-append result opcode) elem code lstack lindex)
    (begin
      (display opcode)
      (fmt/in result elem code lstack lindex)))))

