;;;
;;;  optim.scm: optimizer for the BE's language
;;;
;;;
;;;  Copyright (C) 2001 Samuel Hocevar
;;;  $Id: optim.scm,v 1.3 2001/11/27 18:24:57 sam Exp $
;;; 
;;;  Authors: Samuel Hocevar <sam@zoy.org>
;;; 
;;;  This program is free software; you can redistribute it and/or modify
;;;  it under the terms of the GNU General Public License as published by
;;;  the Free Software Foundation; either version 2 of the License, or
;;;  (at your option) any later version.
;;;  
;;;  This program is distributed in the hope that it will be useful,
;;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;  GNU General Public License for more details.
;;; 
;;;  You should have received a copy of the GNU General Public License
;;;  along with this program; if not, write to the Free Software
;;;  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.


;;;
;;;  simplify-tree: recursively simplify a syntaxic tree
;;;

(define simplify-tree (lambda (tree)
  (case (car tree)
    ; whole program
    ((prog)
     (cons 'prog (simplify-strsep (cons (cadr tree) (caddr tree)))))
    ; a ::= b
    ((inst_affect)
       (if (null? (cdaddr (cadr tree)))
         (list 'var_affect
           `(var ,(cadar (cdadr tree)))
           (simplify-tree (cadddr tree)))
         (list 'tab_affect
           `(tab ,(cadar (cdadr tree)) ,@(simplify-tree (cadr (cdadr tree))))
           (simplify-tree (cadddr tree)))))
    ; var ::= ident index
    ((var)
     (if (null? (cdaddr tree))
       (list 'var (cadadr tree))
       (cons 'tab (cons (cadadr tree) (simplify-tree (caddr tree))))))
    ; index ::= [ indices ]
    ((index)
     (simplify-strsep (cons (cadr (caddr tree)) (caddr (caddr tree)))))
    ; either a SI..ALORS..SINON expression, or a simple expression
    ((expr)
     (if (equal? "SI" (cadr tree))
       ; expr ::= SI expr1 oper_comp expr2 ALORS expr3 SINON expr4
       ; ==> (test (oper_comp expr1 expr2) expr3 expr4)
       (list 'test (list (let ((sym (cadr (cadddr tree))))
                           (if (equal? sym "#") '!= (string->symbol sym)))
                         (simplify-tree (caddr tree)) ; expr1
                         (simplify-tree (cadddr (cdr tree)))) ; expr2
                   (simplify-tree (cadddr (cdddr tree))) ; expr3
                   (simplify-tree (cadddr (cddddr (cdr tree))))) ; expr4
       ; expr ::= expr_simple
       (simplify-tree (cadr tree))))
    ; simple expression
    ((expr_simple)
     (simplify_opsep (cons (cadr tree) (caddr tree))))
    ; tete_expr : use simplify_opsep to simplify
    ((tete_expr)
     (cond
       ((list? (cadr tree))
        ; tete_expr ::= terme
        (simplify-tree (cadr tree)))
       ((equal? "+" (cadr tree))
        ; tete_expr ::= + terme
        (simplify-tree (caddr tree)))
       (else
        ; tete_expr ::= - terme
        (list 'neg (simplify-tree (caddr tree))))))
    ; terme : use simplify_opsep to simplify
    ((terme)
     (simplify_opsep (cons (cadr tree) (caddr tree))))
    ; fact : either (expr) or var
    ((fact)
     (if (equal? "(" (cadr tree))
       ; fact ::= ( expr )
       (simplify-tree (caddr tree))
       ; fact ::= var
       (simplify-tree (cadr tree))))
    (else
     (dtc-error "simplifier found an unknown symbol "
                "(" (symbol->string (car tree)) ")")
     (list "***")))))

; simplify an operator-separated list ( foo oper foo ... oper foo )
(define simplify_opsep (lambda (l)
  (simplify_opsep/in (cdr l) (simplify-tree (car l)))))

(define simplify_opsep/in (lambda (l result)
  (if (null? l)
    result
    (simplify_opsep/in
      (cddr l)
      (list (string->symbol (cadar l)) result (simplify-tree (cadr l)))))))

; simplify a string-separated list ( foo "blah" foo "blah" ... "blah" foo )
(define simplify-strsep (lambda (l)
  (if (null? l)
    ()
    (if (string? (car l))
      (simplify-strsep (cdr l))
      (cons (simplify-tree (car l)) (simplify-strsep (cdr l)))))))

;;;
;;;  optimize-tree: recursively optimize a simplified tree
;;;

(define optimize-tree (lambda (tree)
  (case (car tree)
    ; Leaf - can't be optimized
    ((var)
     (list (car tree) (cadr tree)))
    ; Structural stuff - recursively optimize
    ((prog var_affect tab_affect)
     (cons (car tree) (optimize-list (cdr tree))))
    ((tab)
     (cons (car tree) (cons (cadr tree) (optimize-list (cddr tree)))))
    ((test)
     (cons (car tree)
           (cons (optimize-tree (cadr tree))
                 (optimize-list (cddr tree)))))
    ; Now this can be directly optimized because it's symmetrical
    ((!= = *)
     (if (equal? 'var (caadr tree))
       (cons (car tree) (list (optimize-tree (caddr tree)) (cadr tree)))
       (cons (car tree) (optimize-list (cdr tree)))))
    ((+)
     (cond
       ((equal? 'neg (caadr tree)) ; -a+b ==> b-a
        (optimize-tree (list '- (caddr tree) (cadadr tree))))
       ((equal? 'neg (caaddr tree)) ; a+-b ==> a-b
        (optimize-tree (list '- (cadadr tree) (caddr tree))))
       ((equal? 'var (caadr tree))
        (cons (car tree) (list (optimize-tree (caddr tree)) (cadr tree))))
       (else
        (cons (car tree) (optimize-list (cdr tree))))))
    ; This can be optimized because it's reversible
    ((>)
     (if (equal? 'var (caadr tree))
       (cons '< (list (optimize-tree (caddr tree)) (cadr tree)))
       (cons (car tree) (optimize-list (cdr tree)))))
    ((<)
     (if (equal? 'var (caadr tree))
       (cons '> (list (optimize-tree (caddr tree)) (cadr tree)))
       (cons (car tree) (optimize-list (cdr tree)))))
    ((>=)
     (if (equal? 'var (caadr tree))
       (cons '<= (list (optimize-tree (caddr tree)) (cadr tree)))
       (cons (car tree) (optimize-list (cdr tree)))))
    ((<=)
     (if (equal? 'var (caadr tree))
       (cons '>= (list (optimize-tree (caddr tree)) (cadr tree)))
       (cons (car tree) (optimize-list (cdr tree)))))
    ; We can try to optimize this
    ((neg)
     (cond 
       ((equal? '- (caadr tree)) ; -(a-b) ==> (b-a)
        (optimize-tree (list '- (caddr (cadr tree)) (cadr (cadr tree)))))
       ((equal? 'neg (caadr tree)) ; -(-x) ==> x
        (optimize-tree (cadadr tree)))
       (else
        (list (car tree) (optimize-tree (cadr tree))))))
    ((-)
     (cond
       ((equal? 'neg (caadr tree)) ; -a-b ==> -(a+b)
        (list 'neg (optimize-tree (list '+ (caddr tree) (cadadr tree)))))
       ((equal? 'neg (caaddr tree)) ; a--b ==> a+b
        (optimize-tree (list '+ (caddr tree) (cadadr tree))))
       (else
        (cons (car tree) (optimize-list (cdr tree))))))
    ; This can't be easily optimized (or can it ?)
    ((/)
     (cons (car tree) (optimize-list (cdr tree))))
    (else
     (dtc-error "optimizer found an unknown symbol "
                "(" (symbol->string (car tree)) ")")
     tree))))

; recursively optimize a list of things to optimize
(define optimize-list (lambda (l)
  (if (null? l)
    ()
    (cons (optimize-tree (car l)) (optimize-list (cdr l))))))

