;;;
;;;  language.scm: grammar and everything for a simple calculus language
;;;
;;;
;;;  Copyright (C) 2001 Samuel Hocevar
;;;  $Id: language.scm,v 1.1 2001/11/28 00:55:02 sam Exp $
;;; 
;;;  Authors: Samuel Hocevar <sam@zoy.org>
;;; 
;;;  This program is free software; you can redistribute it and/or modify
;;;  it under the terms of the GNU General Public License as published by
;;;  the Free Software Foundation; either version 2 of the License, or
;;;  (at your option) any later version.
;;;  
;;;  This program is distributed in the hope that it will be useful,
;;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;  GNU General Public License for more details.
;;; 
;;;  You should have received a copy of the GNU General Public License
;;;  along with this program; if not, write to the Free Software
;;;  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.


;;;
;;;  The grammar
;;;

(define *grammar*
  '((expr        (tete_expr (* (oper_add terme))))
    (tete_expr   (terme)
                 ("+" terme)
                 ("-" terme))
    (oper_add    ("+") ("-"))
    (terme       (fact (* (oper_mul fact))))
    (oper_mul    ("*") ("/"))
    (fact        ("(" expr ")")
                 (number))
    (number      (digit (* (digit))))
    (digit       ("0") ("1") ("2") ("3") ("4")
                 ("5") ("6") ("7") ("8") ("9"))))

;;;
;;;  simplify-tree: recursively simplify a syntaxic tree
;;;

(define simplify-tree (lambda (tree)
  (case (car tree)
    ; simple expression
    ((expr)
     (simplify_opsep (cons (cadr tree) (caddr tree))))
    ; tete_expr : use simplify_opsep to simplify
    ((tete_expr)
     (cond
       ((list? (cadr tree))
        ; tete_expr ::= terme
        (simplify-tree (cadr tree)))
       ((equal? "+" (cadr tree))
        ; tete_expr ::= + terme
        (simplify-tree (caddr tree)))
       (else
        ; tete_expr ::= - terme
        (list '- (simplify-tree (caddr tree))))))
    ; terme : use simplify_opsep to simplify
    ((terme)
     (simplify_opsep (cons (cadr tree) (caddr tree))))
    ; fact : either (expr) or number
    ((fact)
     (if (equal? "(" (cadr tree))
       ; fact ::= ( expr )
       (simplify-tree (caddr tree))
       ; fact ::= number
       (simplify-tree (cadr tree))))
    ; number : we just convert it to a real number
    ((number)
     (digits->number 0 (cons (cadr tree) (caddr tree))))
    (else
     (dtc-error "simplifier found an unknown symbol "
                "(" (symbol->string (car tree)) ")")
     (list "***")))))

; transform a list of digits into a real number
(define digits->number (lambda (n d)
  (if (null? d)
    n
    (digits->number (+ (* 10 n) (string->number (cadar d))) (cdr d)))))

; simplify an operator-separated list ( foo oper foo ... oper foo )
(define simplify_opsep (lambda (l)
  (simplify_opsep/in (cdr l) (simplify-tree (car l)))))

(define simplify_opsep/in (lambda (l result)
  (if (null? l)
    result
    (simplify_opsep/in
      (cddr l)
      (list (string->symbol (cadar l)) result (simplify-tree (cadr l)))))))

;;;
;;;  optimize-tree: recursively optimize a simplified tree
;;;  generate-pseudocode: generate assembly code from a syntax tree
;;;  format-pseudocode: nicely format pseudocode
;;;  output-pseudocode: nicely print pseudocode to stdout
;;;

(define optimize-tree (lambda (tree) tree))
(define generate-pseudocode eval)
(define format-pseudocode number->string)
(define output-pseudocode (lambda (code) (display code) (newline)))

